import collections.abc
from collections import defaultdict
from collections.abc import MutableMapping
from typing import Any


def flatten(d: MutableMapping[Any, Any], parent_key: str = "", sep: str = ".") -> tuple[list[str], list[Any]]:
    items: list[str] = []
    values: list[Any] = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, MutableMapping):
            inner_items, inner_values = flatten(v, new_key, sep=sep)
            items.extend(inner_items)
            values.extend(inner_values)
        else:
            items.append(new_key)
            values.append(v)
    return items, values


def recursive_update(d, u):
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = recursive_update(d.get(k, {}), v)
        else:
            d[k] = v
    return d


class InfiniteDict(defaultdict):
    def __init__(self):
        defaultdict.__init__(self, self.__class__)

    def expand(self, levels: list[str], value: Any):
        current = self
        for level in levels[:-1]:
            current = current[level]

        current[levels[-1]] = value



from pydantic import BaseModel


def _create_dict(schema_head: dict, definitions: dict):
    schema_dict = {}
    if "type" in schema_head and schema_head["type"] == "object":
        for key, value in schema_head["properties"].items():
            if "$ref" in value:
                sub_schema = value["$ref"].replace("#/$defs/", "")
                schema_dict[key] = _create_dict(definitions[sub_schema], definitions)
            else:
                schema_dict[key] = None
    else:
        return None
    return schema_dict


def get_schema(config_cls: type[BaseModel]):
    schema = config_cls.model_json_schema()
    definitions = schema["$defs"] if "$defs" in schema else {}
    return _create_dict(schema, definitions)



import re
from pathlib import Path
from typing import Any


def interpolate_paths(config: dict[str, Any], config_dir: Path) -> dict[str, Any]:
    """Recursively interpolate path placeholders in the configuration dictionary.

    Supports the following placeholders:
    - %(here)s: Replaced with the absolute path to the directory containing the config file

    :param config: Configuration dictionary to process
    :param config_dir: Absolute path to the directory containing the configuration file
    :return: Configuration dictionary with interpolated paths
    """
    return _recursive_interpolate(config, config_dir)


def _recursive_interpolate(obj: Any, config_dir: Path) -> Any:
    """Recursively process configuration values and interpolate path placeholders.

    :param obj: Any configuration value (dict, list, str, or other)
    :param config_dir: Absolute path to the directory containing the configuration file
    :return: Processed value with interpolated paths
    """
    if isinstance(obj, dict):
        return {key: _recursive_interpolate(value, config_dir) for key, value in obj.items()}
    elif isinstance(obj, list):
        return [_recursive_interpolate(item, config_dir) for item in obj]
    elif isinstance(obj, str):
        return _interpolate_string(obj, config_dir)
    else:
        return obj


def _interpolate_string(value: str, config_dir: Path) -> str:
    """Interpolate path placeholders in a string value.

    :param value: String value that may contain placeholders
    :param config_dir: Absolute path to the directory containing the configuration file
    :return: String with placeholders replaced
    """
    # Define available placeholders
    placeholders = {
        "here": str(config_dir.resolve()),
    }

    # Replace %(name)s style placeholders
    pattern = re.compile(r"%\((\w+)\)s")

    def replace_match(match: re.Match[str]) -> str:
        placeholder_name = match.group(1)
        if placeholder_name in placeholders:
            return placeholders[placeholder_name]
        # If placeholder is not recognized, leave it unchanged
        return match.group(0)

    return pattern.sub(replace_match, value)



# module of all supported annotations
import os
from typing import Any

import pydantic
from pydantic import Field


def EnvField(default: Any, *, env_var: str, **kwargs: Any):
    return Field(default_factory=lambda: os.environ.get(env_var, default), **kwargs)


def Secret(env_var: str):
    return EnvField(..., env_var=env_var)


def OpenRange(gt: float | None = None, lt: float | None = None):
    return Field(..., gt=gt, lt=lt)


def ClosedRange(ge: float | None = None, le: float | None = None):
    return Field(..., ge=ge, le=le)


def MixedRange(
    gt: float | None = None,
    ge: float | None = None,
    lt: float | None = None,
    le: float | None = None,
):
    return Field(..., gt=gt, ge=ge, lt=lt, le=le)


__all__ = list(pydantic.__all__) + [Secret, OpenRange, ClosedRange, MixedRange]  # pyright: ignore[reportUnsupportedDunderAll]



"""base module for all file backends e.g. yaml, json, xml, ..."""

from abc import abstractmethod
from typing import Any, TextIO


class BaseFileParser:
    """Base class for all file backends"""

    @abstractmethod
    def get_endings(self) -> list[str]:
        """Returns a list of file endings (including point e.g. .json)
        supported by the given backend
        :return: List of supported file endings
        """
        pass

    @abstractmethod
    def parse(self, file: TextIO) -> dict[str, Any]:
        """Base method for converting the given string into a python dict
        :param file: TextIO with access to the raw file data
        :return: Dict of file content
        """
        pass



import argparse

from pydantic import BaseModel

from confme.utils.dict_util import InfiniteDict, flatten
from confme.utils.typing import get_schema


def argument_overwrite(config_cls: type[BaseModel]) -> InfiniteDict:
    # extract possible parameters
    config_dict = get_schema(config_cls)
    parameters: list[str] = []
    if config_dict is not None:
        parameters, _ = flatten(config_dict)

    # get arguments from command line
    parser = argparse.ArgumentParser(prefix_chars="+/")
    group = parser.add_argument_group(
        "Configuration Parameters",
        "With the parameters specified bellow, the configuration values from the config file can be overwritten.",
    )
    for param in parameters:
        group.add_argument(f"++{param}", required=False)
    args, unknown = parser.parse_known_args()

    # find passed arguments and fill it into the dict structure
    infinite_dict = InfiniteDict()
    for param in parameters:
        value = getattr(args, param)
        if value:
            infinite_dict.expand(param.split("."), value)

    return infinite_dict



import functools
import inspect
import warnings


def log_deprecated(msg: str):
    warnings.simplefilter("always", DeprecationWarning)
    warnings.warn(msg, category=DeprecationWarning, stacklevel=2)
    warnings.simplefilter("default", DeprecationWarning)


def deprecated(reason: str = ""):
    def decorator(func):
        if inspect.isclass(func):
            old_init = func.__init__

            @functools.wraps(func.__init__)
            def new_init(*args, **kwargs):
                log_deprecated(f"Call to deprecated class {func.__name__} ({reason}).")
                old_init(*args, **kwargs)

            func.__init__ = new_init
            return func
        else:

            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                log_deprecated(f"Call to deprecated class {func.__name__} ({reason}).")
                return func(*args, **kwargs)

            return wrapper

    return decorator



import logging
import os
from pathlib import Path
from typing import Any, Callable, ClassVar

from pydantic import BaseModel
from tabulate import tabulate
from typing_extensions import Self

from confme import source_backend
from confme.core.argument_overwrite import argument_overwrite
from confme.core.env_overwrite import env_overwrite
from confme.utils.base_exception import ConfmeException
from confme.utils.dict_util import flatten, recursive_update


class BaseConfig(BaseModel):
    __KEY_LOOKUP__: ClassVar[list[str]] = ["env", "environment", "environ", "stage"]
    __config_path__: ClassVar[Path | None] = None
    __default_env__: ClassVar[str | None] = None
    __cache__: ClassVar[dict[str, "BaseConfig"]] = {}

    @classmethod
    def load(cls, path: Path | str) -> Self:
        """Load your configuration file into your config class structure.
        :param config_class: Root class to map the configuration file to
        :param path: path to configuration file
        :return: instance of config_class with all values added from the config file
        """
        config_content = source_backend.parse_file(path)
        config_content = recursive_update(config_content, env_overwrite(cls))
        config_content = recursive_update(config_content, argument_overwrite(cls))

        return cls.model_validate(config_content)

    @classmethod
    def load_from_dict(cls, config_content: dict[str, Any]) -> Self:
        config_content = recursive_update(config_content, env_overwrite(cls))
        config_content = recursive_update(config_content, argument_overwrite(cls))

        return cls.model_validate(config_content)

    @classmethod
    def register_folder(
        cls,
        config_folder: Path,
        default_env: str | None = None,
        strict: bool = False,
    ) -> None:
        """Register a folder where configuration files are drawn based on the environment.
        :param config_folder: Path to the folder with configuration files per environment
        :param default_env: Default environment that should be used if none is specified via environment variable
        :param strict: If True, an exception is raised if no configuration file is found that exactly matches env name.
        """
        cls.__config_path__ = config_folder
        cls.__default_env__ = default_env
        cls._strict = strict
        cls.__cache__ = {}

    @classmethod
    def get(cls) -> Self:
        """Get the corresponding configuration based on the environment. Thereby, the configuration class is loaded
        once and cached for subsequent calls.
        :return: instance of config_class with all values added from the config file
        """
        env = cls._get_current_env()
        if env not in cls.__cache__:
            cls.__cache__[env] = cls._load_file(env)

        return cls.__cache__[env]  # type: ignore[return-value]

    @classmethod
    def _load_file(cls, environment: str) -> Self:
        if cls.__config_path__ is None:
            raise ConfmeException("Config path not set. Call register_folder() first.")
        files = list(cls.__config_path__.glob(pattern="*"))
        if cls._strict:
            selected_files = [f for f in files if f.name == environment or f.stem == environment]
        else:
            selected_files = [f for f in files if environment in f.name]

        if len(selected_files) <= 0:
            raise ConfmeException(f"No configuration found for environment {environment} in files {files}")
        elif len(selected_files) > 1:
            logging.warning(
                f"More than one file found matching environment {environment} in"
                f"files {files}. Using file {selected_files[0]}"
            )
            file = selected_files[0]
        else:
            file = selected_files[0]

        return cls.load(file)

    @classmethod
    def _get_current_env(cls) -> str:
        keys = [key for key in os.environ.keys() if key.lower() in cls.__KEY_LOOKUP__]

        if len(keys) <= 0 and cls.__default_env__ is None:
            raise Exception(
                "You're using the register_folder / get combination one of the "
                "following environment variables need to be set: ENV, ENVIRONMENT,"
                "ENVIRON, env, environment, environ"
            )
        elif len(keys) <= 0 and cls.__default_env__ is not None:
            return cls.__default_env__
        elif len(keys) > 1:
            logging.warning(f"More than one environment variable set using the value of {keys[0]}")
            key = keys[0]
        else:
            key = keys[0]

        return os.environ[key].lower()

    def update_by_str(self, path: str, value: Any) -> None:
        """Given a path string separated by dots (.) this method allows to update nested configuration objects.
        CAVEAT: For this update no type check is applied!
        e.g. given this configuration
        ```
        class DatabaseConfig(BaseConfig):
            host: str
            port: int
            user: str

        class MyConfig(BaseConfig):
            name: int
            database: DatabaseConfig

        config = MyConfig.load('test.yaml')

        config.update_by_str('database.host', 'my new host')
        :param path: dot (.) separated string which value should be updated
        :param value: update value
        """
        path_parts = path.split(".")
        current = self
        for i, segment in enumerate(path_parts):
            if not hasattr(current, segment):
                raise ConfmeException(f"{segment} not found in path {'.'.join(path_parts[:i])}!")

            # for the last item we don't want to assign it to the current element because we want to assign the given
            # value instead
            if i + 1 < len(path_parts):
                current = getattr(current, segment)
            else:
                setattr(current, segment, value)

    def get_flat_repr(self) -> list[tuple[str, Any]]:
        """Returns a flat representation of your configuration structure (tree).
        e.g. given this configuration
        ```
        class DatabaseConfig(BaseConfig):
            host: str
            port: int
            user: str

        class MyConfig(BaseConfig):
            name: int
            database: DatabaseConfig

        config = MyConfig.load('test.yaml')
        config.get_flat_repr()
        ...
        [('name', 'test1'),
        ('database.host', 'localhost'),
        ('database.port', '5678'),
        ('database.user', 'db_user')]

        :return: list of configuration parameter key and it's corresponding value
        """
        keys, values = flatten(self.model_dump())
        return list(zip(keys, values))

    def log_config(self, print_fn: Callable[[str], None] = logging.info):
        """Prints/logs the configuration in a flat format.
        :param print_fn: print callable to overwrite can be used e.g. with log_config(print_fn=print)
        """
        flat_config = self.get_flat_repr()
        str_config = tabulate(flat_config, headers=["Key", "Value"], tablefmt="github")
        print_fn(str_config)



from os import path
from pathlib import Path
from typing import Any

from confme.source_backend.backend_yaml import YamlFileParser
from confme.utils.path_interpolation import interpolate_paths

FILE_PARSER = [YamlFileParser()]


def parse_file(file_path: str | Path, interpolate: bool = True) -> dict[str, Any]:
    """Parses the given file with the right file parser based on the filename ending of the
    given file_path. Supports path interpolation with %(here)s placeholder.

    :param file_path: path to the file
    :param interpolate: If True, interpolate path placeholders like %(here)s. Defaults to True.
    :return: Dict with content of the file
    """
    file_path_obj = Path(file_path)
    file_path_str = str(file_path_obj)
    ending = path.splitext(file_path_str)[-1]
    applicable_file_parsers = [p for p in FILE_PARSER if ending in p.get_endings()]
    if len(applicable_file_parsers) < 0:
        raise Exception(f"File Ending {ending} not known")
    if len(applicable_file_parsers) > 1:
        raise Exception("More than one parser registered for this file ending... 🧐")

    with open(file_path_str) as file:
        config = applicable_file_parsers[0].parse(file)

    if interpolate:
        config_dir = file_path_obj.parent
        config = interpolate_paths(config, config_dir)

    return config



"""module for parsing yaml files"""

import logging
from typing import Any, TextIO

import yaml
from yaml.parser import ParserError

from confme.source_backend.backend_base import BaseFileParser


class YamlFileParser(BaseFileParser):
    """File Parser for yaml files"""

    def get_endings(self) -> list[str]:
        """Returns all yaml file endings
        :return: List of yaml file endings
        """
        return [".yaml", ".yml"]

    def parse(self, file: TextIO) -> dict[str, Any]:
        """Converts the given yaml file into a python dict
        :param file: yaml file stream
        :return: Content of yaml file converted to dict
        """
        try:
            return yaml.safe_load(file)
        except ParserError as err:
            logging.exception("Not able to parse yaml file")
            raise ParserError from err
